
#include <stdio.h>
#include <sys/time.h>

#include <appclass.h>
#include <dlist.h>
 
#ifdef TRACE_MEM
#include <tracemem.h>
#endif
        

/***************************************************************************
 *
 * DList(AppClass) Methods : A doubly linked list.
 *
 ***************************************************************************/

static DList *dlist_new_node      ( AppClass *data );
static void   dlist_construct_node( DList *node, AppClass *data );
static void   dlist_destroy_node  ( void *node );


static DList *dlist_new_node( AppClass *data )
{
    DList *node;

    node = app_new( DList, 1 );
    dlist_construct_node( node, data );
    app_class_overload_destroy( (AppClass *)node, dlist_destroy_node );

    return node;
}

static void dlist_construct_node( DList *node, AppClass *data )
{
   app_class_construct( (AppClass *) node );

   node->prev = NULL;
   node->next = NULL;

   node->data = data;
}

static void dlist_destroy_node( void *node )
{
   DList *_node = (DList *)node;

   if (_node == NULL) {
      return;
   }

   app_class_unref( _node->data );
   
   app_class_destroy( node );
}

/** \brief Add a new node to the end of the list.
 *
 * If cmp argument is not NULL, use cmp() to see if node already exists and
 * don't add node if it exists.
 *
 * This function can unref data if not added, if asked by user (*res != 0).
 * The res variable is set to 1 if data is not added.
 * 
 * head is modified only when the list is empty.
 *  So it is safe to add node to an existing list.
 */

DList *dlist_add( DList *head, AppClass *data, DListFP_Cmp cmp, int *res )
{
   DList *node = head;

   if (head == NULL) {
      /* The list is empty, make new node the head. */
      if ( res ) {
         *res = 0; /* elem added to the list */
      }
      return dlist_new_node( data );
   }

   /* Walk the list to find the end */

   while (node) {
      if ( cmp && ((*cmp)(node->data, data) == 0) ) {
	 /* node already exists and we were asked to keep nodes unique */
	 if ( res ) { 
	    if ( *res ) { 
	       app_class_unref( data ); /* elem destroyed */
	    }
	    *res = 1;  /* elem not added to the list */
	 }
	 break;
      }

      if (node->next == NULL) {
	 /* at the tail */
	 node->next = dlist_new_node( data );
	 node->next->prev = node;
         if ( res ) {
             *res = 0; /* elem added to the list */
         }
	 break;
      }
      
      /* move on to next node */
      node = node->next;
   }

   return head;
}

/** \brief Add a new node at the head of the list. */

DList *dlist_add_head( DList *head, AppClass *data )
{
   DList *node = dlist_new_node( data );

   if (head) {
      head->prev = node;
      node->next = head;
   }

   return node;
}

/** \brief Conditionally delete a node from the list.
 *
 *  Delete a node from the list if the node's data matches the specified
 *  data. Returns the head of the modified list.
 */

DList *dlist_delete( DList *head, AppClass *data, DListFP_Cmp cmp )
{
   DList *node = head;

   if (cmp == NULL) {
      msg_fatal( "compare function not specified" );
   }

   while (node) {
      if ( (*cmp)(node->data, data) == 0 ) {
	 if ( (node->prev == NULL) && (node->next == NULL) ) {
	    /* deleting only node in list (node is head and tail) */
	    head = NULL;
	 } else if (node->prev == NULL) {
	    /* node is head, but other nodes exist */
	    node->next->prev = NULL;
	    head = node->next;
	 } else if (node->next == NULL) {
	    /* node is tail, but other nodes exist */
	    node->prev->next = NULL;
	 } else {
	    /* node is not head nor tail */
	    node->prev->next = node->next;
	    node->next->prev = node->prev;
	 }

	 /* this will also unref the node->data */
	 app_class_unref( (AppClass *)node );
	 
	 return head;
      }
      
      /* move on to next node */
      node = node->next;
   }
   
   /* if we get here, data wasn't found, just return original head */
   return head;
}

/** \brief Blow away the entire list.
 */

void dlist_delete_all( DList *head )
{
   DList *node;
   
   while (head) {
      node = head;
      head = head->next;
      
      app_class_unref( (AppClass *)node );
   }
}

/** \brief Lookup an item in the list.
 *
 *  Walk the list pointed to by head and return a pointer to the data if
 *  found. If not found, return NULL. 
 *
 *  \param head The head of the list to be iterated.
 *  \param data The data to be passed to the func when it is applied.
 *  \param cmp  A function to be used for comparing the items.
 *
 *  \return     A pointer to the data found, or NULL if not found.
 */

AppClass *dlist_lookup( DList *head, AppClass *data, DListFP_Cmp cmp )
{
   DList *node = head;

   if (cmp == NULL) {
      msg_fatal( "compare function not specified" );
   }
   
   while (node) {
      if ( (*cmp)(node->data, data) == 0 ) {
	 return node->data;
      }
      node = node->next;
   }
   
   /* If we get here, no node was found, return NULL. */

   return NULL;
}

/*
 * a simple node compare function for the iterator.
 */

int dlist_iterator_cmp( AppClass *n1, AppClass *n2 )
{
    /*
     * Since this is only used in the iterator, we are guaranteed that it is
     * safe to compare data pointers because both n1 and n2 came from the
     * list.
     */

   return (int)(n1 - n2);
}

/** \brief Iterate over all elements of the list.
 *
 *  For each element, call the user supplied iterator function and pass it the
 *  node data and the user_data. If the iterator function return non-zero,
 *  remove the node from the list.
 *
 *  \param head The head of the list to be iterated.
 *  \param func The function to be applied.
 *  \param user_data The data to be passed to the func when it is applied.
 *
 *  \return A pointer to the head of the possibly modified list.
 */

DList *dlist_iterator( DList *head, DListFP_Iter func, void *user_data )
{
   DList *node = head;
   AppClass *data;

   if (func == NULL) {
      msg_fatal( "no iteration func supplied" );
   }

   while (node) {
      data = node->data;
      node = node->next; /* must be before remove node */
      if ( (*func)(data, user_data) ){
	 /* remove node */
	 head = dlist_delete( head, data, dlist_iterator_cmp );
      }
   }
   
   return head;
}

/*
 * return number of nodes in list
 */
int dlist_get_nelem (DList *head )
{
   DList *node = head;
   int n = 0;

   while (node) {
      n++;
      node = node->next; /* must be before remove node */
   }
   return n;
}

/*
 * get a member of the list at index
 */
AppClass *dlist_get_ndata(DList *head, int n)
{
   DList *node = head;
   int i = 0;
   
   while (node) {
      if ( i == n ) {
         return node->data;
      }
      i++;
      node = node->next;
   }
   return NULL;
}

